/*
 * Licensed Materials - Property of IBM
 *
 * com.ibm.etools.c
 *
 *  Copyright IBM Corporation 2004. All Rights Reserved. 
 *
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.etools.c.importer;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.ibm.etools.c.CArray;
import com.ibm.etools.c.CClassifier;
import com.ibm.etools.c.CDatatype;
import com.ibm.etools.c.CDerivableType;
import com.ibm.etools.c.CDerived;
import com.ibm.etools.c.CField;
import com.ibm.etools.c.CFunction;
import com.ibm.etools.c.CParameter;
import com.ibm.etools.c.CPointer;
import com.ibm.etools.c.CStruct;
import com.ibm.etools.c.CStructured;
import com.ibm.etools.c.CTypedElement;
import com.ibm.etools.c.CTypedef;
import com.ibm.etools.c.CUnion;
import com.ibm.etools.c.datatypes.CChar;
import com.ibm.etools.c.datatypes.CEnumeration;
import com.ibm.etools.c.datatypes.CSignedChar;
import com.ibm.etools.c.datatypes.CUnsignedChar;
import com.ibm.etools.c.datatypes.CUnsignedShort;
import com.ibm.etools.c.datatypes.CWchar;

/**
 * The CQuery class provides convenient methods to query C model objects.
 */
public class CQuery {

    public static final String copyright = "Licensed Material - Property of IBM\n com.ibm.etools.c\n(C) Copyright IBM Corp. 2004 - All Rights Reserved.\nUS Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp."; //$NON-NLS-1$

    private static CQuery _instance = new CQuery();

    public CQuery() {
    }

    /**
     * Return the single instance
     */
    public static CQuery instance() {
        return _instance;
    }

    /**
     * Return the number of children for a container type (struct, union, enum
     * and function)
     */
    public static int numberOfChildren(Object type) {

        int size = 0;
        if (type instanceof CStructured) {
            size = numberOfFields((CStructured) type);
        } else if (type instanceof CEnumeration) {
            size = numberOfEnumerators((CEnumeration) type);
        } else if (type instanceof CFunction) {
            size = numberOfParameters((CFunction) type);
        }

        return size;
    }

    /**
     * Return the number of fields within a struct or union
     */
    public static int numberOfFields(CStructured astructured) {

        List fields = astructured.getContains();
        if (fields == null) {
            return 0;
        } else {
            return fields.size();
        }
    }

    /**
     * Return the number of enumerators within an enumeration
     */
    public static int numberOfEnumerators(CEnumeration anenum) {

        List enumerators = anenum.getEnumerator();
        if (enumerators == null) {
            return 0;
        } else {
            return enumerators.size();
        }

    }

    /**
     * Return the number of parameters within a function (not including the
     * return parameter)
     */
    public static int numberOfParameters(CFunction afunction) {

        List parameters = afunction.getParameter();
        if (parameters == null || parameters.size() == 0) {
            return 0;
        } else {
            if (hasReturnType(afunction)) {
                return parameters.size() - 1;
            } else {
                return parameters.size();
            }
        }
    }

    /**
     * Return the type of a typed element
     */
    public static CClassifier typeOf(CTypedElement element) {
        CClassifier type = element.getType();
        if (type != null)
            return type;
        else {
            CResource.writeLog(CQuery.instance(), "typeOf", CResource.getString("_ERROR_Type_Not_Found", element.getName()));
            return null;
        }
    }

    /**
     * Return the base type of a typed element
     */
    public static CClassifier baseTypeOf(CTypedElement element) {
        return baseType(typeOf(element));
    }

    /**
     * Is this an array type?
     */
    public static boolean isArray(CClassifier type) {
        return (type instanceof CArray);
    }

    /**
     * Is this a char array?
     */
    public static boolean isCharArray(CClassifier type) {
        if (type instanceof CTypedef)
            type = typedefBaseType((CTypedef) type);

        if (type instanceof CArray) {
            CArray array = (CArray) type;
            if (array.getDerives() instanceof CClassifier) {
                CClassifier referenceType = (CClassifier) array.getDerives();
                if (referenceType instanceof CTypedef)
                    referenceType = typedefBaseType((CTypedef) referenceType);

                return ((referenceType instanceof CChar) && !(referenceType instanceof CUnsignedChar) && !(referenceType instanceof CSignedChar));
            } else
                return false;
        } else
            return false;
    }

    /**
     * Is this a wchar_t array?
     */
    public static boolean isWcharArray(CClassifier type) {
        if (type instanceof CTypedef)
            type = typedefBaseType((CTypedef) type);

        if (type instanceof CArray) {
            CArray array = (CArray) type;
            CDerivableType referenceType = array.getDerives();
            if (referenceType instanceof CTypedef)
                return (typedefBaseType((CTypedef) referenceType) instanceof CWchar);
            else
                return (referenceType instanceof CWchar);
        } else
            return false;
    }

    /**
     * Is this an unsigned short array?
     */
    //@P1A - entire method
    public static boolean isUnsignedShortArray(CClassifier type) {
        if (type instanceof CTypedef)
            type = typedefBaseType((CTypedef) type);

        if (type instanceof CArray) {
            CArray array = (CArray) type;
            if (array.getDerives() instanceof CClassifier) {
                CClassifier referenceType = (CClassifier) array.getDerives();
                if (referenceType instanceof CTypedef)
                    referenceType = typedefBaseType((CTypedef) referenceType);

                return (referenceType instanceof CUnsignedShort);
            } else
                return false;
        } else
            return false;
    }

    /**
     * Is this a pointer type?
     */
    public static boolean isPointer(CClassifier type) {
        return (type instanceof CPointer);
    }

    /**
     * Is this a C primitive type?
     */
    public static boolean isPrimitiveType(CClassifier type) {
        return (type instanceof CDatatype);
    }

    /**
     * Is this the first field of a struct/union?
     */
    public static boolean isFirstField(CField afield) {
        CStructured astructured = afield.getContainer();
        if (astructured == null)
            return false;
        else {
            List fields = astructured.getContains();
            return (fields.get(0) == afield);
        }
    }

    /**
     * Return the number of dimensions for an array.
     */
    public static int arrayDimensionSize(CArray array) {
        int size = 0;
        Object baseType = array;
        while (baseType instanceof CArray) {
            CArray myArray = (CArray) baseType;
            baseType = myArray.getDerives();
            size = size + 1;
        }
        return size;
    }

    /**
     * Return the base type of an array. The base type is not an array type.
     */
    public static CClassifier arrayBaseType(CArray array) {

        Object baseType = array;
        while (baseType instanceof CArray) {
            CArray myArray = (CArray) baseType;
            baseType = myArray.getDerives();
        }

        if (baseType instanceof CClassifier) {
            return (CClassifier) baseType;
        } else
            return null;

    }

    /**
     * Return the base type of a typedef.
     */
    public static CClassifier typedefBaseType(CTypedef atypedef) {
        Object baseType = atypedef;
        while (baseType instanceof CTypedef) {
            CTypedef myTypedef = (CTypedef) baseType;
            baseType = myTypedef.getDerives();
        }

        if (baseType instanceof CClassifier) {
            return (CClassifier) baseType;
        } else
            return null;
    }

    /**
     * Return the base type of a more complex type
     */
    public static CClassifier baseType(CClassifier type) {

        CClassifier baseType = type;
        if (baseType instanceof CTypedef) {
            baseType = typedefBaseType((CTypedef) baseType);
            return baseType(baseType);
        } else if (baseType instanceof CArray) {
            baseType = arrayBaseType((CArray) baseType);
            return baseType(baseType);
        } else {
            return baseType;
        }
    }

    /**
     * Is this an aggregate type?
     */
    public static boolean isAggregate(CClassifier type) {

        CClassifier basetype = baseType(type);
        return (basetype instanceof CStruct || basetype instanceof CUnion);
    }

    public static boolean isAggregate(CTypedElement element) {
        return isAggregate(typeOf(element));
    }

    /**
     * Does this function have a declared return type?
     */
    public static boolean hasReturnType(CFunction afunction) {
        List parameters = afunction.getParameter();
        Iterator iterator = parameters.iterator();
        while (iterator.hasNext()) {
            Object child = iterator.next();
            if (child instanceof CParameter) {
                CParameter aparameter = (CParameter) child;
                if (aparameter.getDirection().getName().equals("return"))
                    return true;
            }
        }
        return false;
    }

    /**
     * Is this type an unresolved type?
     * 
     * A struct or union is unresolved if it contains no fields. A derived type
     * is unresolved if the deriving type is unknown.
     */
    public static boolean isUnresolvedType(CClassifier type) {
        return isUnresolvedType(type, false);
    }

    /**
     * Is this type an unresolved type?
     * 
     * By default, we only search one level to determine resolveness. If the
     * recursive flag is true, we will do a recursive search. If any type that
     * this type references to is unresolved, then this type is also considered
     * to be unresolved.
     */
    public static boolean isUnresolvedType(CClassifier type, boolean recursive) {

        if (type instanceof CDerived) {
            return isUnresolvedDerived((CDerived) type, recursive);
        } else if (type instanceof CStructured) {
            return isUnresolvedStructured((CStructured) type, recursive);
        }
        // All primitive types are resolvable.
        else {
            return false;
        }

    }

    /**
     * Is this derived type unresolved? If the recursive flag is false, we only
     * check the direct diriving type. If it is true, we will trace back
     * recursively. If any of the parent types is unresolved, then this type is
     * also considered to be unresolved.
     */
    public static boolean isUnresolvedDerived(CDerived aderived, boolean recursive) {

        CDerivableType derivingType = aderived.getDerives();
        if (derivingType == null) {
            return true;
        } else if (recursive) {
            if (derivingType instanceof CFunction) {
                return false;
            } else if (derivingType instanceof CClassifier) {
                return isUnresolvedType((CClassifier) derivingType, recursive);
            } else {
                return true;
            }
        } else {
            return false;
        }
    }

    /**
     * Is this Structured (struct or union) unresolved? A struct or union is
     * unresolved if it does not contain any fields. If the recursive flag is
     * true, we also look at the types of the fields. If any of the fields has
     * an unresolved type, the the whole structured is also considered to be
     * unresolved.
     */
    public static boolean isUnresolvedStructured(CStructured astructured, boolean recursive) {

        List fields = astructured.getContains();
        if (fields == null || fields.size() == 0) {
            return true;
        } else if (recursive) {
            Iterator it = fields.iterator();
            while (it.hasNext()) {
                Object child = it.next();
                if (child instanceof CField) {
                    CField afield = (CField) child;
                    CClassifier type = afield.getType();
                    if (type != null) {
                        // Avoid infinite recursion when the field's type is a
                        // pointer to the same structured.
                        //
                        if ((type instanceof CPointer) && ((CPointer) type).getDerives() == astructured) {
                            continue;
                        } else {
                            if (isUnresolvedType(type, recursive))
                                return true;
                        }
                    } else {
                        return true;
                    }
                } else if (child instanceof CStructured) {
                    if (isUnresolvedStructured((CStructured) child, recursive))
                        return true;
                }
            }
            return false;
        } else {
            return false;
        }
    }

    /**
     * Return the list of unresolved types from a given object list
     */
    public static ArrayList getUnresolvedTypes(List objectList, boolean recursive) {

        ArrayList list = new ArrayList();
        Iterator it = objectList.iterator();
        while (it.hasNext()) {
            Object child = it.next();
            if (child instanceof CClassifier) {
                if (isUnresolvedType((CClassifier) child, recursive))
                    list.add(child);
            }
        }
        return list;
    }

    /**
     * Return the list of unresolved types from a given object list
     */
    public static ArrayList getUnresolvedTypes(List objectList) {
        return getUnresolvedTypes(objectList, false);
    }

}